{******************************************************************************}
{******************************************************************************}
{** This Delphi component was developed by Andreas Hard, 1998-05-02          **}
{** All rights belongs to Andreas Hard and the code may NOT be altered       **}
{** without permission.                                                      **}
{** THE TABLECHANGEDNOTIFY ONLY WORKS WITH PARADOX TABLES                    **}
{** PLEASE OBSERVERE! The TableChangedNotify is a little "buggy" so don't    **}
{** use It. Try instead to develop It ;-)                                    **}
{******************************************************************************}
{******************************************************************************}
unit PackTable;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Db, DBTables, DbiTypes, BDE, extctrls;
type
  TOnTableChange = procedure of object;
  TPackTable     = class(TTable)
  private
    FChangeTableNotify : boolean;
    FOnTableChange     : TOnTableChange;
    FRegistered        : boolean;
    FCloneCur          : hDBICur;
    FCheckInterval     : word;
    pDummyBuf          : pBYTE;
    Timer              : TTimer;
    FValueOnStartup    : boolean;
  protected
    procedure PackError(Error : DBIResult); virtual;
    procedure SetTableNotify(Value: boolean); virtual;
    procedure SetCheckInterval(Value: word); virtual;
    procedure UnregisterCallback; virtual;
    procedure RegisterCallback; virtual;
    procedure InternalClose; override;
    procedure Loaded; override;
    procedure CheckForChange(Sender: TObject);
  public
    TableProperties: CURProps;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Pack;
  published
    property OnTableChange:
             TOnTableChange read FOnTableChange write FOnTableChange;
    property TableCheckInterval:
             word read FCheckInterval write SetCheckInterval;
    property TableChangeNotify:
             boolean read FChangeTableNotify write SetTableNotify;
  end;

procedure Register;

implementation

uses
  DbiProcs, DbiErrs;

// -------------------------------------- Function called by DbiRegisterCallBack
function Callback(ecbType : CBType; iClientData : LongInt;
                  pCbInfo : pCBInputDesc) : CBRType; stdcall;
begin
  Result := cbrUSEDEF;                                 // Return default type...
                      // If type = cbTABLECHANGED, drop FOnTableChanged event...
  if ecbtype = cbTABLECHANGED then
    begin
                   // iClientData is really a pointer to It self (TPackTable)...
      if iClientData <> 0 then
                     // If FOnTableChange event has a value, "fire the event"...
        if Assigned(TPackTable(iClientData).FOnTableChange) then
           TPackTable(iClientData).FOnTableChange;
    end;
end;

constructor TPackTable.Create(AOwner: TComponent);
begin
  inherited;
  FCloneCur := nil; pDummyBuf := nil;
  FillChar(TableProperties, sizeof(TableProperties), 0);
  FRegistered := False;
                                                           // Create a TTimer...
  Timer := TTimer.Create(nil);
  Timer.Enabled := False;
                                                           // Default = 5 sek...
  Timer.Interval := 5000;
  FCheckInterval := 5000;
                                                      // Check if Design-mode...
  if not (csDesigning in ComponentState) then
    Timer.OnTimer := CheckForChange;
end;

destructor TPackTable.Destroy;
begin
                                                               // Free TTimer...
  Timer.Free;
  inherited;
end;

procedure TPackTable.Loaded;
begin
  inherited;
  SetTableNotify(FValueOnStartup);
end;

procedure TPackTable.CheckForChange(Sender: TObject);
var
  rslt: DBIResult;
begin
  DbiCheckRefresh;
  rslt := DbiGetRecord(FCloneCur, dbiNOLOCK, pDummyBuf, nil);
  if rslt <> DBIERR_NONE then
    begin
      rslt := DbiGetNextRecord(FCloneCur, dbiNOLOCK, pDummyBuf, nil);
      if rslt <> DBIERR_NONE then
        begin
          Check(DbiSetToEnd(FCloneCur));
          DbiGetPriorRecord(FCloneCur, dbiNOLOCK, pDummyBuf, nil);
          DbiGetRecord(FCloneCur, dbiNOLOCK, pDummyBuf, nil);
        end;
    end;
end;

procedure TPackTable.SetCheckInterval(Value: word);
begin
  FCheckInterval := Value;
  Timer.Interval := Value;
end;

procedure TPackTable.SetTableNotify(Value: boolean);
begin
  if (csReading in ComponentState) then
    begin
      FValueOnStartup := Value;
      Exit;
    end;

  if Value = True then
    begin
      if Active = False then
        raise EDatabaseError.Create(
              'The table must be opened to be able to register "callbacks"');
      RegisterCallback;
    end
  else
    if Active = True then
      UnregisterCallback;
  FChangeTableNotify := Value;
end;

procedure TPackTable.InternalClose;
begin
  UnregisterCallback;
  FChangeTableNotify := False;
  inherited;
end;

procedure TPackTable.UnregisterCallback;
begin
  if FRegistered = True then
    begin
      Check(DbiRegisterCallBack(FCloneCur, cbTABLECHANGED, 0,
                                sizeof(CBInputDesc), nil , nil));
      Check(DbiCloseCursor(FCloneCur));
      FreeMem(pDummyBuf);
      FRegistered := False;
      Timer.Enabled := False;
    end;
end;

procedure TPackTable.RegisterCallback;
begin
  if FRegistered = False then
    begin
      Check(DbiGetCursorProps(Handle, TableProperties));
      if TableProperties.szTableType <> szPARADOX then
        raise EDatabaseError.Create('Ony works with Paradox-tables');
      pDummyBuf := AllocMem(TableProperties.iRecBufSize);
      Check(DbiCloneCursor(Handle, True, False, FCloneCur));
      Check(DbiSetToBegin(FCloneCur));
      Check(DbiGetNextRecord(FCloneCur, dbiNOLOCK, nil, nil));
      Check(DbiRegisterCallBack(FCloneCur, cbTABLECHANGED, longint(self),
                                                            0, nil, @Callback));
      FRegistered := True;
      if not (csDesigning in ComponentState) then
        Timer.Enabled := True;
    end;
end;

procedure TPackTable.Pack;
var
  OldExclusiveState,
  OldActiveState            : Boolean;
  TableDescription          : pCRTblDesc;
  TempResult                : DBIResult;
  TempDbHandle              : hDBIdb;
  ExtType                   : TTableType;
  TableNameExt              : String;
begin
  OldExclusiveState := Exclusive;
  OldActiveState    := Active;

  DisableControls;

  if Active and (not Exclusive) then
    Close;
  if (not Exclusive) then
    Exclusive := True;
  if (not Active) then
    Open;
  TempDbHandle := DBHandle;

  TableNameExt := UpperCase(ExtractFileExt(TableName));
  if TableNameExt = '.DBF' then
    ExtType := ttdBase
  else
    if ((TableNameExt = '.DB') or (TableNameExt = '')) then
      ExtType := ttParadox;

  TempResult := DBIERR_NOTSUPPORTED;
  if (TableType = ttdBase) or
     ((TableType = ttDefault) and
     (ExtType = ttdBase)) then
    begin
      TempResult := DbiPackTable(DBHandle, Handle, nil, nil, True);
    end
  else
    if (TableType = ttParadox) or
       ((TableType = ttDefault) and
       (ExtType = ttParadox)) then
      begin
        New(TableDescription);
        FillChar(TableDescription^, SizeOf(TableDescription^), 0);

        With TableDescription^ do
        begin
          bPack := True;
          StrPCopy(szTblName, TableName);
          StrPCopy(szTblType, szParadox);
        end;

        Close;
        TempResult := DbiDoRestructure(TempDBHandle, 1,
                      TableDescription, nil, nil, nil, False);

        Dispose(TableDescription);
      end;

  if TempResult <> DBIERR_NONE then
    PackError(TempResult);

  Close;

  Exclusive := OldExclusiveState;
  Active    := OldActiveState;

  EnableControls;
end;

procedure TPackTable.Packerror(Error : DBIResult);
var
  ErrorString : DBIMSG;
begin
  DbiGetErrorString(Error, ErrorString);
  MessageDlg(ErrorString, mtError, [mbOK], 0);
end;

procedure Register;
begin
  RegisterComponents('BugWash', [TPackTable]);
end;

end.
